/***************************************************************************
                          ADM_mpegout.cpp  -  description
                             -------------------
    begin                : Sat Oct 12 2002
    copyright            : (C) 2002 by mean
    email                : fixounet@free.fr
    
    Here we allocate the buffer for libmpeg2dec
    We have to look inside it because libmpeg2 will pop out the images
    		once it reaches 3 images in the buffer
      
      image[0]= Backward prediction 
      image[1]= Forward prediction 
      image[2]= Current (B frame mainly) 
      
    
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#ifdef HAVE_SYS_PARAM_H
#include <sys/param.h>
#endif
#ifdef __FreeBSD__
          #include <sys/types.h>
#endif
#include "config.h"
#include "ADM_library/default.h"
extern "C"
{
#include "libMpeg2Dec/video_out.h"
#include "libMpeg2Dec/mpeg2.h"
#include "libMpeg2Dec/mm_accel.h"

};

#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <math.h>


#include "avifmt.h"
#include "avifmt2.h"

#include "ADM_editor/ADM_Video.h"
#include "ADM_audio/aviaudio.hxx"

#include "ADM_library/fourcc.h"
#include "ADM_mpeg2dec/ADM_mpegscan.h"
#include "libMpeg2Dec/mpeg2_internal.h"
#include "ADM_toolkit/toolkit.hxx"


// avoid exporting our stuff...
typedef struct yv12_frame_s {
    vo_frame_t 	vo;   
    int			reclaimed;
}yv12_frame_t;

typedef struct yv12_instance_s 
{
    vo_instance_t 	vo;
    int 						prediction_index;
    vo_frame_t * 		frame_ptr[3];
    yv12_frame_t 		frame[3];
    mpeg2decHeader *father;

} yv12_instance_t;


uint8_t *yv12_get_backward( void );

static void 	yv12_close (vo_instance_t * _instance);
static int 		yv12_setup (vo_instance_t * _instance, int width, int height);
static int 		yv12_alloc_frames (yv12_instance_t * instance, int width, int height);
static void 	yv12_draw_frame (vo_frame_t * _frame);
static vo_frame_t * yv12_get_frame (vo_instance_t * _instance, int flags);

//static uint8_t *zero=NULL;
mpeg2decHeader *_father=NULL;

static    uint8_t *base;
static    uint8_t *base2;

static 		uint8_t *last=NULL; //last buffer used
//static 		uint8_t _force=0;

void mpeg2decHeader::setFather( mpeg2decHeader *f)
{
		_father=f;
}

vo_instance_t * vo_yv12_open (void)
{
    yv12_instance_t * instance;
  

    instance = new yv12_instance_t;
    if (instance == NULL)
				return NULL;
    memset(instance,0,sizeof(*instance));

    instance->vo.setup = yv12_setup;
    instance->vo.close = yv12_close;
    instance->vo.get_frame = yv12_get_frame;
    instance->father=_father;
    
	   printf("\n open called");

    return (vo_instance_t *) instance;
}

int yv12_setup (vo_instance_t * _instance, int width, int height)
{
    yv12_instance_t * instance;

    instance = (yv12_instance_t *) _instance;
    
     instance->prediction_index = 1;

     yv12_alloc_frames (instance,  width,  height);
	   printf("\n setup called for %d x %d pic \n",width,height);
    return 0;
}
void yv12_close (vo_instance_t * _instance)
{
   UNUSED_ARG(_instance);
   printf("\n close called");
}
#ifdef HAVE_MEMALIGN
/* some systems have memalign() but no declaration for it */
void * memalign (size_t align, size_t size);
#else
/* assume malloc alignment is sufficient */
#define memalign(align,size) malloc (size)
#endif

int yv12_alloc_frames (yv12_instance_t * instance, int width, int height)
{
    int i;
    uint32_t qsize;
    
    qsize=(width*height)>>2;
    for (i = 0; i < 3; i++) 
    {
			instance->frame_ptr[i] = (vo_frame_t *) (instance->frame + i);
			base=(uint8_t *)memalign(16,qsize*6);
			assert(base);

			instance->frame[i].vo.base[0] = base;
			instance->frame[i].vo.base[1] = base+qsize*5;
			instance->frame[i].vo.base[2] = base+qsize*4;
			instance->frame[i].vo.copy = NULL;
			instance->frame[i].vo.field = NULL;
			instance->frame[i].vo.draw = yv12_draw_frame;
			instance->frame[i].vo.instance = (vo_instance_t *) instance;
			instance->frame[i].reclaimed=0;
    }
    base=instance->frame[0].vo.base[0];
    base2=instance->frame[0].vo.base[0];

    printf("\n Alloc frame called");
    return 0;
}

//extern mpeg2dec_t mpeg2dec;
vo_frame_t * yv12_get_frame (vo_instance_t * _instance, int flags)
{
//static int fr=0;
//		int cfr;
     
     	yv12_instance_t * instance;
    	instance = (yv12_instance_t *)_instance;
     
	    if (flags & VO_PREDICTION_FLAG) 
  	  {
 						instance->prediction_index ^= 1;
            // we force a I frame here
            // 
//						printf("using innner buf : %d\n",instance->prediction_index);
						last=	instance->frame_ptr[instance->prediction_index]->base[0];					
						instance->frame[instance->prediction_index].reclaimed=0;
						return instance->frame_ptr[instance->prediction_index];
    	} 
//					printf("using inner buf : 2\n");
     				instance->frame[2].reclaimed=0;
				return instance->frame_ptr[2];
}
    
uint8_t *yv12_get_backward( void )
{		
		
	
		return  last; //last; //instance->frame_ptr[0];
}

// Call back from decoder to avidemux when a frame is ready    
void yv12_draw_frame (vo_frame_t * _frame)
{
				// retrieve father
				mpeg2decHeader *father;
				yv12_instance_t *inst;
				
				inst= (yv12_instance_t *)(_frame->instance);
				
				father=inst->father;
//				printf("\n Draw!\n");
				father->push(_frame->base[0]);
				inst->frame[0].reclaimed=1;
}
